# Copyright (C) 2005 by Ryan Lovett <ryan@stat.berkeley.edu>

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

import os, struct, platform
import StringIO, zipfile, gzip

ACCT_LINUX_FMT = 'c3HI9HI16s12x'
ACCT_SUNOS_FMT = '2cIiIi6H8s'

class AcctFile:
    def __init__ (self, filename):
        self.filename = filename
        self.processes = []
        self.AcctProcess = AcctProcess

    def getFH (self):
        return open (self.filename)

    def parse (self):
        f = self.getFH ()

        rec = f.read (self.ACCT_SIZE)
        while rec != '':
            self.processes.append (self.AcctProcess (rec))
            rec = f.read (self.ACCT_SIZE)

        f.close ()

class LinuxAcctFile (AcctFile):
    def __init__ (self, filename):
        AcctFile.__init__ (self, filename)
        self.ACCT_FMT = ACCT_LINUX_FMT
        self.ACCT_SIZE = struct.calcsize (self.ACCT_FMT)
        self.AcctProcess = LinuxAcctProcess

class GzipLinuxAcctFile (LinuxAcctFile):
    def getFH (self):
        return gzip.GzipFile (self.filename, 'r')

class SunOSAcctFile (AcctFile):
    def __init__ (self, filename):
        AcctFile.__init__ (self, filename)
        self.ACCT_FMT = ACCT_SUNOS_FMT
        self.ACCT_SIZE = struct.calcsize (self.ACCT_FMT)
        self.AcctProcess = SunOSAcctProcess

class ZipSunOSAcctFile (SunOSAcctFile):
    def getFH (self):
        f = zipfile.ZipFile (self.filename)
        a = f.read ('var/adm/pacct')
        return StringIO.StringIO (a)

class GzipSunOSAcctFile (SunOSAcctFile):
    def getFH (self):
        return gzip.GzipFile (self.filename, 'r')

class AcctProcess:
    def toSeconds (self, value):
        '''Convert input from the kernel's accounting units to
        seconds.'''
        pass

    def decode_comp_t (self, value):
        exp = (value >> 13) & 07    # 3 bit base-8 exponent
        retval = value & 017777        # 13 bit fraction
        while exp > 0:
            # cast to long to silence v2.3 warnings
            retval = long (retval << 3)
            exp -= 1

        return retval

    def encode_comp_t (self, value):
        '''From linux/kernel/acct.c'''
        MANTSIZE = 13   # 13 bit mantissa
        EXPSIZE = 3     # Base 8 (3 bit) exponent
        MAXFRACT = (1 << MANTSIZE) - 1  # Maximum fractional value
 
        exp = rnd = 0
 
        while value > MAXFRACT:
            rnd = value & ((1 << (EXPSIZE - 1)))    # Round up?
            value >>= EXPSIZE
            exp += 1
 
        value += 1
        if rnd != 0 and value > MAXFRACT:
            value >>= EXPSIZE
            exp += 1
 
        exp <<= MANTSIZE
        exp += value
 
        return exp

class LinuxAcctProcess (AcctProcess):
    def __init__ (self, rec):
        self.ACCT_FMT = ACCT_LINUX_FMT
        self.HZ = 100

        entry = struct.unpack (self.ACCT_FMT, rec)

        (self.flag, self.uid, self.gid, self.tty,
        self.btime, self.utime, self.stime, self.etime,
        self.mem, self.io, self.rw, self.minflt,
        self.majflt, self.swaps, self.exitcode, self.comm) \
        = entry

    def expand (self, value):
        return self.decode_comp_t (value)

    def toSeconds (self, value):
        '''Convert input from the kernel's accounting units to
        seconds.'''
        # unit = 1/USER_HZ of a second = 1/100th of a second
        return (value * 1.0 / self.HZ)

class SunOSAcctProcess (AcctProcess):
    def __init__ (self, rec):
        self.ACCT_FMT = ACCT_SUNOS_FMT
        self.HZ = 100
        entry = struct.unpack (self.ACCT_FMT, rec)

        (self.flag, self.stat, self.uid, self.gid,
        self.tty, self.btime, self.utime, self.stime,
        self.etime, self.mem, self.io, self.rw, self.comm) \
        = entry[0:12] + (entry[12].replace ('\x00', ''),)

    def expand (self, value):
        return self.decode_comp_t (value)

    def toSeconds (self, value):
        '''Convert input from the kernel's accounting units to
        seconds.'''
        # unit = 1/USER_HZ of a second = 1/100th of a second
        return (value * 1.0 / self.HZ)

def load (filename):
    system = platform.system ()

    if system == 'Linux':
        acct_file = LinuxAcctFile (filename)
    elif system == 'SunOS':
        acct_file = SunOSAcctFile (filename)

    return acct_file
